//  Gandalf's Staff - by Aaron Haras 2018/2019
//  Based on the orginal SnowGlobe, and subsiquent PoolThingy versions.
// 
//  Concepts/Features                 //    Orginal code adapted from......
//  -----------------                       ---------------------------------
//  Colour Lights                     // Strandtest variant from the NeoPixel library
//  Multitasking                      // Multitasking Arduino part 3 and 4
//  State Machine                     // From BlinkWithoutDelay, Multitasking Arduino and OneButton.
//  DoubleClick                       // http://www.mathertel.de/Arduino/OneButtonLibrary.aspx
//  MusicMaker                        // Adafruit's MusicMaker Featherwing.
//  BlueTooth                         // my own hacks
// 
//
//           Table of Contents for code (rough)
// 
//  Library declarations                                    - line 105
//  Global Variable declarations                            - line 111
//    - OneButton variables                                 - line 118
//    - Hardware variables                                  - line 124
//    - Misc global variables                               - line 138 
//    - Instantiate sound device                            - line 156
//  Main Class declaration                                  - line 160
//    - class variables (local)                             - line 164
//    Methods/Functions of the class
//    - constructor()                                       - line 184
//    - Update()      ***                                   - line 199
//    - Increment()                                         - line 242
//    - SlowDown()                                          - line 271
//    - SpeedUp()                                           - line 279
//    - Reverse()                                           - line 290
//    - NextPattern()    ***                                - line 308
//    - Sparks(interval)                                    - line 385
//    - SparksUpdate()                                      - line 401
//    - RainbowCycle(interval,direction)                    - line 427
//    - RainbowCycleUpdate()                                - line 438
//    - TheaterChase(color,color,interval,direction)        - line 452
//    - TheaterChaseUpdate()                                - line 464
//    - ColorWipe(color,interval,direction)                 - line 482
//    - ColorWipeUpdate()                                   - line 493
//    - Scanner(color,Interval)                             - line 501
//    - ScannerUpdate()                                     - line 510
//    - Fade(color,color,steps,interval,direction)          - line 535
//    - FadeUpdate()                                        - line 547
//    - DimColor(color)                                     - line 561
//    - DimColor(color,factor)                              - line 568
//    - ColorSet(color)                                     - line 582
//    - Red(color)                                          - line 592
//    - Green(color)                                        - line 598
//    - Blue(color)                                         - line 604
//    - Wheel(position)                                     - line 611
//    - WhiteLight()                                        - line 629
//    - WhiteLightUpdate                                    - line 635
//    - LightningBolt()                                     - line 640
//    - Thunder()                                           - line 647
//    - DetectEvil(interval)                                - line 653
//    - DetectEvilUpdate()                                  - line 674
//    - More_Evil()                                         - line 689
//    - Less_Evil()                                         - line 694
//    - More_Bright()                                       - line 699
//    - Less_Bright()                                       - line 708
//    - More_Volume()                                       - line 718
//    - Less_Volume()                                       - line 725
//    - Next_Colour()                                       - line 731
//    - Next_Colour(int ColourID)                           - line 738 - Play evil line 765, play good line 799
//  End of main NeoPattern Class                            - line 833
//  Main Declarations                                       - line 836
//    - Ring1Complete()                                     - line 838
//    - NeoPattern object declaration                       - line 841
//  Setup()                                                 - line 845 - attaches callbacks for buttons
//  Loop()                                                  - line 925
//  Other classes                                           -
//    PrintDirectory()                                      - line 944
//    OneButton callbacks                                   - line 980
//      - Function Switch Events                            - line 990
//        - click()                                         - line 992
//        - doubleclick()                                   - line 996
//        - longPressStart()                                - line 1001
//        - longPress()                                     - line 1005
//        - longPressStop()                                 - line 1010
//      - Toggle Switch Events                              - line 1029
//        - click()                                         - line 1031
//        - doubleclick()                                   - line 1067
//        - longPressStart()                                - line 1074
//        - longPressStop()                                 - line 1078
//        - longPress()                                     - line 1086
//      - Increment Switch Events                           - line 1093
//        - click()                                         - line 1095
//        - doubleclick()                                   - line 1131
//        - longPressStart()                                - line 1166
//        - longPressStop()                                 - line 1199
//        - longPress()                                     - line 1234
//      - Decrement Switch Events                           - line 1240
//        - click()                                         - line 1242
//        - doubleclick()                                   - line 1278
//        - longPressStart()                                - line 1310
//        - longPressStop()                                 - line 1339
//        - longPress()                                     - line 1381
//  CheckPlayer()                                           - line 1392
//  GetNextSong(file)                                       - line 1405
//  PlayError(int)                                          - line 1431
//  FlashCurrentPattern()                                   - line 1446


// Here are the libraries needed to run this code.
#include <Adafruit_NeoPixel.h>                    // Need this! Tells compiler to add the code for Neo_Pixels
#include <SPI.h>                                  // Communication to the sound device
#include <SD.h>                                   // SD card files
#include <Adafruit_VS1053.h>                      // MusicMaker Lbrary
#include <OneButton.h>                            // files are in the arduino/library/onebutton sub folder.

// Variable Declarations
//
//  OneButton Variables. Each switch (A1-A4) declared below has multiple click events that
// are attached upon declaration. The callback methods for each switch are all at the end of the code
// There is a Click, DoubleClick, LongPress, LongPressStart, and LongPressStop events available.
// Declaration of the OneButton variable accepts ( PinNumber, pull internal resistor ________)
// Pins A1-A3 attach to reed switch and then to ground.
OneButton btn_Function(A1, true);                 // Switch to change Function(pattern) using Arduino pin A1.           - 
OneButton btn_Toggle(A2, true);                   // Switch to toggle something.(same pattern differnt colour,etc)      - 
OneButton btn_Increment(A4, true);                // used to increment something (brightness, speed,etc)                - 
OneButton btn_Decrement(A3, true);                // the decrement Switch.                                              - 

// ProTrinket Pins 
byte Flashlight_Pin = A0;                         // Pin to fire the transistor/relay for centre FlashLight.                                       - wire is Brown/White
int NeoPixil_Signal_Pin = 10;                     // Pin for the data to the NeoPixel Ring                              - wire is Yellow/Blue
int Bluetooth_Pin = 9;                            // Pin to power (on/off)  BlueTooth Transmitter
int MOSI_Pin = 11;                                // MOSI for the Music Maker
int MISO_Pin = 12;                                // MISO for the Music Maker
int SCK_Pin = 13;                                 // SCK for the Music Maker
int VS1053_CS_Pin = 8;                            // MP3CS for the Music Maker
int VS1053_DCS_Pin = 6;                           // XDS for the Music Maker
int CARDCS_Pin = 4;                               // SDSC for the Music Maker
int VS1053_DREQ_Pin = 3;                          // DReq for the Music Maker - ideally an intrrupt pin.
#define VS1053_RESET   -1                         // VS1053 reset pin (not used!)

enum  direction { FORWARD, REVERSE };             // Pattern directions supported 
int maxBrightness = 240;                          // max brightness the pixel ring will be allowed to achieve.250 max
int Brightness = 200;                             // overall brightness of the LED ring
int Interval_Max = 1000;                          // Maximmum the Interval duration can be set to.
int Interval_Min = 1;                             // Minimum the Interval duration can be set to
int Interval_Increment;                           // Amount to increment/decrement the Interval
int CurrentPattern = 1;                           // The Current Pattern, replaced enumerations from original code
int CurrentColour = 1;                            // current colour. List of colours to cycle through
int Evil_R = 125;                                 // Value of the red component for Good/Evil Colours
int Evil_G = 125;                                 // Value of the green component
int Evil_B = 125;                                 // Value of the blue component
int Evil_Inc = 0;                                 // Amoumt to Increment/Decrement the Evil Colours
int Current_Evil_MP3=1;                           // Index to cycle a list of Evil Sounds
int Current_Good_MP3=1;                           // Index for the Good list of songs
int Volume = 20;                                  // Starting volume for Music Maker, (Note: Smaller = louder)
int Min_Volume = 30;                              // Set Min Volume value

// Declare a variable/object called "musicPlayer" that is of type "Adafruit_VS1053_FilePlayer".
// Detatils of using the musicmaker Featherwing can be found at _________________________

Adafruit_VS1053_FilePlayer musicPlayer = 
  Adafruit_VS1053_FilePlayer(VS1053_RESET, VS1053_CS_Pin, VS1053_DCS_Pin, VS1053_DREQ_Pin, CARDCS_Pin);

//
class NeoPatterns : public Adafruit_NeoPixel      // NeoPattern Class - derived from the Adafruit_NeoPixel class
{
  public:

    // Variables that are members of the class. 
    direction Direction;                          // direction to run the pattern
    unsigned long Interval;                       // milliseconds between updates
    unsigned long lastUpdate;                     // last update of position
    uint32_t Color1, Color2;                      // Patterns often fade between colours.
    uint16_t TotalSteps;                          // Total number of steps in the pattern
    uint16_t Index;                               // Current step within the pattern
    int SparkPosition;                            // Position of Random Spark for Sparks method
    int SparkIntensity;                           // Brightness of the sparks pattern
    int SparkIncAmount;                           // Increment/decrement for sparks pattern
    void (*OnComplete)();                         // Callback on completion of pattern
    
    // Constructor for the class
    // When initializing (constructing) the class the user needs to pass
    // along the following information.
    // - The size of the NeoPixel Ring(strip,etc). This is the number of pixels,
    // - The pin number of the arduino board.
    // - Parameters for the type of Neo_pixel Ring, and finally
    // - routine to go to on Callback. - MoreDetailNeeded -
    //
    NeoPatterns(uint16_t pixels, uint8_t pin, uint8_t type, void (*callback)())
      : Adafruit_NeoPixel(pixels, pin, type)
    {
      OnComplete = callback;
    }

    // Update()
    // The Update method of the class is the workhorse. Here the code will use the
    // microprocessor's clock to determine if it is time to do something. By comparing
    // the elapsed time against a pattern's Interval (time between updates), This 
    // allows the application to share the processor's time between parts of code. 
    // For example: multiple pixel rings, multiple button detection, sound device, etc.
    // The Update method also determines the current pattern and executes 
    // the appropriate Update method for that pattern.
    //
    void Update()
    {
      // get current time in milliseconds.                                // millis()
      // subtract the time the update function was last called.           // millis()-lastUpdate
      // compare the elapsed time against Interval.                       // (millis() - lastUpdate) > Interval
      // If the elapsed time is less than Interval, not enought time has passed, exit update() 
      // If the elapsed time is greater than Interval, execute the rest of th update function.      
      if ((millis() - lastUpdate) > Interval)   
      {
        // It is time to do something        
        lastUpdate = millis();                // reset the time of the last update    // might move this to end of update function
        switch (CurrentPattern)               // Use the switch statement to act on the current pattern
        {
          case 1:
            WhiteLightUpdate();                // call the update method of the WhiteLight pattern
            break;
          case 2:
            RainbowCycleUpdate();             // call the update method of the RainbowCycle pattern
            break;
          case 3:
            FadeUpdate();                     // call the update method of the Fade pattern
            break;
          case 4:
            SparksUpdate();                   // call the update method of the Sparks pattern
            break;
          case 5:
            DetectEvilUpdate();               //   call the update method of the DetectEvil pattern
            break;
          case 6:                             // Couple of extras for new patterns                      
            break;
          case 7:
            break;
          case 8:
            break;
          case 9:
            break;
          default:
            break;
        }     // end switch
      }       // end test elapsed time against Interval
    }         // End update method 

    // Method to Increment the Index and reset at the end
    void Increment()
    {
      if (Direction == FORWARD)
      {
        Index++;
        if (Index >= TotalSteps)
        {
          Index = 0;
          if (OnComplete != NULL)
          {
            OnComplete(); // call the comlpetion callback
          }
        }
      }
      else // Direction == REVERSE
      {
        --Index;
        if (Index <= 0)
        {
          Index = TotalSteps - 1;
          if (OnComplete != NULL)
          {
            OnComplete(); // call the comlpetion callback
          }
        }
      }
    }       // End Method Increment

    // Method to SlowDown to speed the pattern is moving
    void SlowDown()
    {
      Interval = Interval + Interval_Increment; // increasing the interval reduces refresh rate and thus slows down pattern
      if (Interval > Interval_Max)
      {
        Interval = Interval_Max;
      }
    }
    // Method to speed up the current pattern
    void SpeedUp()
    {
      Interval = Interval - Interval_Increment;
      if (Interval < Interval_Min)
      {
        Interval = Interval_Min;
      }
    }
    
    // Method to Reverse the current pattern's direction
    void Reverse()
    {
      if (Direction == FORWARD)
      {
        Direction = REVERSE;
        Index = TotalSteps - 1;
      }
      else
      {
        Direction = FORWARD;
        Index = 0;
      }
    }
    
    // NextPattern() 
    // This method switches to the next pattern
    // Notes: I have left some blank/extra case statements for additional patterns
    // Also I only used 4 patterns for my staff, so there is a limit so the blank pattern is not run.
    void NextPattern()
    {
      CurrentPattern++;                                   // increment the pattern
      if (CurrentPattern > 5)                             // If pattern number exceeds 4 reset  ***s/b var "CurrentMaximumNumberOfPatterns"
      {
        CurrentPattern = 1;                               // Reset to first pattern
      }
      switch (CurrentPattern)                             // Use the switch to act on the current pattern                    // 
      {
        case 1:
          // This is a basic soft glowing light. Starts as white.
          WhiteLight();                                   // Initialize the pattern
          WhiteLightUpdate();                             // Upadate method for the pattern
          break;
        case 2:
          // swirling rainbow colours (Slow)
          Interval = 20;                                  // set the speed (interval time)
          Interval_Increment = 5;                         // set the amount to increment/decrement
          Interval_Max = 30;                              // Maximmum the Interval duration can goto
          Interval_Min = 1;                               // Minimum the Interval duration can goto
          RainbowCycle(Interval);                         // Initialize Pattern. set speed (update Interval)
          RainbowCycleUpdate();                           // Call update method
          Brightness = 100;                               // adjust the brightness
          setBrightness(Brightness);                      // call function to set brightness
          break;
        case 3: 
          // Fade
          Interval = 500;                                 // set the interval (speed) 
          Interval_Increment = 150;                       // set the amount to increment for this pattern
          Interval_Max = 1500;                            // Maximmum the Interval duration can goto
          Interval_Min = 20;                              // Minimum the Interval duration can goto
          Fade(Color(255,0,0),Color(0,255,0),255,Interval); // accepts 2 colours, #steps and intervalduration
          FadeUpdate();                                   // Call update method
          setBrightness(Brightness);                      // Set the brightness
          break;
        case 4:
          // Sparks, Random lights fade on and off.
          Interval = 5;                                   // set the interval for the Sparks pattern
          Interval_Increment = 1;                         // Increment anmount for the pattern
          Interval_Max = 20;                              // Maximmum the Interval duration can goto
          Interval_Min = 5;                               // Minimum the Interval duration can goto
          Sparks(Interval);                               // Initalize the pattern. accepts update interval as parameter
          SparksUpdate();                                 // call th Update function
          Brightness = 200;                               // Adjust variable for brightness
          setBrightness(Brightness);                      // Set the brightness
          break;
        case 5:
          // Detect Evil Pattern
          // Initialize Function
          // Call the Update function
            // basic Light
            // Hold on Inc - fast fade to Red
            // Hold on Dec - Fast fade to Blue/green
            DetectEvil(5);                                // Initialize DetctEvil pattern
            DetectEvilUpdate();                           // Call the the Update method of the pattern
          break;
        case 6:
          
          break;

        case 7:
          break;

        case 8:
          break;

        case 9:

        default:
          break;
      } // end Switch
    }   // end method NextPattern()


    // Sparks() method
    // This pattern lights up and dims a random pixel
    //
    void Sparks(uint8_t interval)
    {
      Index = 0;                                      // Reset the index
      Interval = interval;                            // assigne the current interval time
      SparkIntensity = 0;                             // Start sparkat black
      TotalSteps = 100;                               // Set the number of steps to full brightness
      SparkIncAmount = 250 / TotalSteps;              // calculate and set the increment amount

      // reset the pixel ring to all black
      for (int x = 0; x < numPixels(); x++)           // Loop throught the pixels and set them black
      {
        setPixelColor(x, 0, 0, 0);                    // calls function to set the pixel ring to one colour.
      }
      show();                                         // The Show() function pushs changes to the NeoPixel ring.
    }
    // The update function of the Sparks pattern
    void SparksUpdate()
    {
      SparkIntensity  = SparkIntensity + SparkIncAmount;;             // Inc/Dec the count for spark brightness
      if (SparkIntensity <= 0 )                                       // check for below minimum
      {
        setPixelColor(SparkPosition, SparkIntensity, SparkIntensity, SparkIntensity);   //calls function to set colours
        show();                                                       // The Show() function pushs changes to the NeoPixel ring.
        // Color is off, get a new location
        SparkPosition = random(12);                                   // new location
        SparkIntensity = 0;                                           // reset intensity
        SparkIncAmount = (-1) * SparkIncAmount;                       // switch from Increment/decrement
      }
      
      if (SparkIntensity >= 255)                                      // check if past maximum
      {
        SparkIncAmount = (-1) * SparkIncAmount;                       // switch from Increment/decrement
        SparkIntensity = 254;                                         // reset intensity
      }
      setPixelColor(SparkPosition, SparkIntensity, SparkIntensity, SparkIntensity);
      show();                                                         // The Show() function pushs changes to the NeoPixel ring.
    }

    // RainbowCycle( Interval, direction
    // Initialize for a RainbowCycle [attern
    //
    //
    void RainbowCycle(uint8_t interval, direction dir = FORWARD)      // accepts interval time and direction as parameters
    {
      Interval = interval;                                            // assigne the current interval time 
      TotalSteps = 255;                                               // reset maximum steps
      Index = 0;                                                      // reset index
      Direction = dir;                                                // Set direction
    }

    // RainbowCycleUpdate()
    // Update funstion of the Rainbow Cycle Pattern
    //
    //
    void RainbowCycleUpdate()
    {
      for (int i = 0; i < numPixels(); i++)                                   // 
      {
        setPixelColor(i, Wheel( ((i * 256 / numPixels()) + Index) & 255 ) );
      }
      show();
      Increment();
    }

    // Initialize for a Theater Chase
    // Currently not used. *************
    //
    void TheaterChase(uint32_t color1, uint32_t color2, uint8_t interval, direction dir = FORWARD)
    {
      // ActivePattern = THEATER_CHASE;
      Interval = interval;
      TotalSteps = numPixels();
      Color1 = color1;
      Color2 = color2;
      Index = 0;
      Direction = dir;
    }

    // Update the Theater Chase Pattern
    void TheaterChaseUpdate()
    {
      for (int i = 0; i < numPixels(); i++)
      {
        if ((i + Index) % 3 == 0)
        {
          setPixelColor(i, Color1);
        }
        else
        {
          setPixelColor(i, Color2);
        }
      }
      show();
      Increment();
    }

    // Initialize for a ColorWipe
    void ColorWipe(uint32_t color, uint8_t interval, direction dir = FORWARD)
    {
      // ActivePattern = COLOR_WIPE;
      Interval = interval;
      TotalSteps = numPixels();
      Color1 = color;
      Index = 0;
      Direction = dir;
    }

    // Update the Color Wipe Pattern
    void ColorWipeUpdate()
    {
      setPixelColor(Index, Color1);
      show();
      Increment();
    }

    // Initialize for a SCANNNER
    void Scanner(uint32_t color1, uint8_t interval)
    {
      Interval = interval;
      TotalSteps = (numPixels() - 1) * 2;
      Color1 = color1;
      Index = 0;
    }

    // Update the Scanner Pattern
    void ScannerUpdate()
    {
      for (int i = 0; i < numPixels(); i++)
      {
        if (i == Index)  // Scan Pixel to the right
        {
          setPixelColor(i, Color1);
        }
        else if (i == TotalSteps - Index) // Scan Pixel to the left
        {
          setPixelColor(i, Color1);
        }

        else // Fading tail
        {
          // gets colour of previously set pixel, dims it.
          setPixelColor(i, DimColor(getPixelColor(i)));
        }
      }
      show();
      Increment();
    }

    // Initialize for a Fade
    // changed interval to unsigned long
    void Fade(uint32_t color1, uint32_t color2, uint16_t steps, unsigned long interval, direction dir = FORWARD)
    {
      //  ActivePattern = FADE;
      Interval = interval;
      TotalSteps = steps;           // unused
      Color1 = color1;              //
      Color2 = color2;              //
      Index = 0;
      Direction = dir;              //
    }

    // Update the Fade Pattern
    void FadeUpdate()
    {
      // Calculate linear interpolation between Color1 and Color2
      // Optimise order of operations to minimize truncation error
      uint8_t red = ((Red(Color1) * (TotalSteps - Index)) + (Red(Color2) * Index)) / TotalSteps;
      uint8_t green = ((Green(Color1) * (TotalSteps - Index)) + (Green(Color2) * Index)) / TotalSteps;
      uint8_t blue = ((Blue(Color1) * (TotalSteps - Index)) + (Blue(Color2) * Index)) / TotalSteps;

      ColorSet ( Wheel(Index) );                  // new version uses the wheel for a smoother loop.
      show();
      Increment();
    }

    // Calculate 50% dimmed version of a color (used by ScannerUpdate)
    uint32_t DimColor(uint32_t color)
    {
      // I dont really understand this.
      // Shift R, G and B components one bit to the right
      uint32_t dimColor = Color(Red(color) >> 1, Green(color) >> 1, Blue(color) >> 1);
      return dimColor;
    }
    uint32_t DimColor(uint32_t color, int factor)
    {
      uint32_t tmpColor;
      //uint32_t tmpColor
      tmpColor = color;
      for (int x = 0; x <= factor; x++)
      {
        // Shift R, G and B components one bit to the right
        tmpColor = DimColor(tmpColor);
      }
      return tmpColor;
    }

    // Set all pixels to a color 
    void ColorSet(uint32_t color)
    {
      for (int i = 0; i < numPixels(); i++)
      {
        setPixelColor(i, color);
      }
      show();
    }

    // Returns the Red component of a 32-bit color
    uint8_t Red(uint32_t color)
    {
      return (color >> 16) & 0xFF;
    }

    // Returns the Green component of a 32-bit color
    uint8_t Green(uint32_t color)
    {
      return (color >> 8) & 0xFF;
    }

    // Returns the Blue component of a 32-bit color
    uint8_t Blue(uint32_t color)
    {
      return color & 0xFF;
    }

    // Input a value 0 to 255 to get a color value.
    // The colours are a transition r - g - b - back to r.
    uint32_t Wheel(byte WheelPos)
    {
      WheelPos = 255 - WheelPos;
      if (WheelPos < 85)
      {
        return Color(255 - WheelPos * 3, 0, WheelPos * 3);
      }
      else if (WheelPos < 170)
      {
        WheelPos -= 85;
        return Color(0, WheelPos * 3, 255 - WheelPos * 3);
      }
      else
      {
        WheelPos -= 170;
        return Color(WheelPos * 3, 255 - WheelPos * 3, 0);
      }
    }
    void WhiteLight()
    { 
      // set colour to white.
      ColorSet(Color(125, 125, 125));                     // White is roughly equal anounts of RGB
      setBrightness(50);                                  //
    }
    void WhiteLightUpdate()
    { // test removing this!
      // colour should be set and not need update
      show();
    }
    void LightningBolt()
    {
      // Method to make lightning flash. Not used yet. Delay borks app. needs to have a interval/
      digitalWrite(4, HIGH);                              //
      delay(1000);                                        //
      digitalWrite(4, LOW);                               //
    }
    void Thunder()
    {
      // Method to activate the sounf device via a relay and trnsistor
      
    }

    void DetectEvil(unsigned long interval)
    {
      // working oct9/19
      setBrightness(maxBrightness);                                // start at full brightness
      
       Interval = interval;
      // set start colour
      Evil_R = 25;
      Evil_G = 75;
      Evil_B = 25;
      // Set brightness
      ColorSet(Color(Evil_R, Evil_G, Evil_B));                     // White is roughly equal anounts of RGB
      // setBrightness(50);                                  //

      // test
      delay(1000); // test to determine if the brightness gets dimmed below.
      setBrightness(maxBrightness);                                // start at full brightness
      
      
    }
    
    void DetectEvilUpdate()
    {
      // Red is Evil (255,0,0)
      // Blue is good. (0,0,255)
      // This update Increments/Decrements colour and re-paint the pixels. 
      // The change in colour is triggered by the button event
      Evil_R = Evil_R + Evil_Inc;                 // 
      if (Evil_R > 255 ){Evil_R = 255;}           // ? Increase overall brightness once max colour is reached?
      if (Evil_R <0){Evil_R = 0;}                 // 
      Evil_B = Evil_B - Evil_Inc;                 // 
      if (Evil_B < 0){Evil_B = 0;}                // 
      if (Evil_B > 255 ){Evil_B = 255;}           // 
      ColorSet(Color(Evil_R,Evil_G,Evil_B));                    // White is roughly equal anounts of RGB
      show();                                           // 
    }
    void More_Evil()
    {
      Evil_Inc = 1;                               // Set increment amount and direction
      Evil_G = 0;                                 // turn off green
    }
    void Less_Evil()
    {
      Evil_Inc = -1;                              // Set increment amount and direction
      Evil_G = 0;                                 // turn off green
    }
    void More_Bright()
    {
      if (Brightness < 50) { Brightness = Brightness + 10;}
      else{Brightness = Brightness + 25;}
    
      if (Brightness > maxBrightness){ Brightness = maxBrightness;}
      setBrightness(Brightness);
    }

    void Less_Bright()
    {
      if (Brightness < 50){Brightness = Brightness - 10;}     // 
      else {Brightness = Brightness - 25;}                    // 
      
      if (Brightness < 0){Brightness = 0;}                    // 
      
      setBrightness(Brightness);                              //
    }

    void More_Volume()
    {
      Volume = Volume - 5;                          // Increase the Volume,
      if (Volume <1){Volume = 1;}                   // Test for max *** Smaller number is louder ***
      musicPlayer.setVolume(Volume,Volume);         // Pass volume to music player
    }
    
    void Less_Volume()
    {
      Volume = Volume + 5;                            // Increase the Volume,
      if (Volume > Min_Volume){Volume = Min_Volume;}                   // Test for max *** Smaller number is louder ***
      musicPlayer.setVolume(Volume,Volume);         // Pass volume to music player
    }
    Next_Colour()
    {
      CurrentColour ++;                              // increment the current colour.
      if (CurrentColour >5) {CurrentColour = 1;}    // test the range
      Next_Colour(CurrentColour);                    // call method to switch colour
    }

    Next_Colour(int NewColour)
    {
      // method to switch to the next colour.
      switch(CurrentColour)
      {
        case 1:     // white
          ColorSet(Color(250,250,250));             
        break;
        case 2:     // Red
          ColorSet(Color(250,0,0));
        break;      
        case 3:     // orange
          ColorSet(Color(250,250,0));
        break;
        case 4:     // Yellow
          ColorSet(Color(125,250,0));
        break;
        case 5:     // Green
          ColorSet(Color(0,255,0));
        break;
        case 6:     // Blue
          ColorSet(Color(0,0,255));
        break;
        default:
        break;
      } // end case      
    } // end method NextColour
    void Play_Evil()
    {
      // cycle through a list of Evil sounding songs. used in Detect Evil mode
      // Should be a list on sd card. Then play list can change without recompile/upload.
        // read file, find current song, get/play next song.
      // for now, we will use named files
      musicPlayer.stopPlaying();
      switch(Current_Evil_MP3)
      {
        case 1:
          musicPlayer.startPlayingFile("Witches.mp3");                            // 
        break;
        case 2:
          musicPlayer.startPlayingFile("Psycho.mp3");                            // 
        break;
        case 3:
          musicPlayer.startPlayingFile("Cat_1.mp3");                            // 
        break;
        case 4:
          musicPlayer.startPlayingFile("im_bad.mp3");                            // 
        break;
        case 5:
          musicPlayer.startPlayingFile("Psycho.mp3");                            // 
        break;
        default:
        break;
      }
      Current_Evil_MP3++;                                             // Increment song #
      if (Current_Evil_MP3>5)                                         // max 5 in case above
      {
        Current_Evil_MP3 = 1;                                         // reset to first 
      }
    } // end Play_Evil

    void Play_Good()
    {
      // cycle through a list of Good sounding songs. used in Detect Evil mode
      // Should be a list on sd card. Then play list can change without recompile/upload.
      // read file, find current song, get/play next song.
      // for now, we will use named files
      musicPlayer.stopPlaying();
      switch(Current_Good_MP3)
      {
        case 1:
          musicPlayer.startPlayingFile("Good_1.mp3");                            // 
        break;
        case 2:
          musicPlayer.startPlayingFile("Good_2.mp3");                            // 
        break;
        case 3:
          musicPlayer.startPlayingFile("Good_3.mp3");                            // 
        break;
        case 4:
          musicPlayer.startPlayingFile("Good_4.mp3");                            // 
        break;
        case 5:
          musicPlayer.startPlayingFile("Good_5.mp3");                            // 
        break;
        default:
        break;
      }
      Current_Good_MP3++;                                             // Increment song #
      if (Current_Good_MP3>5)                                         // max 5 in case above
      {
        Current_Good_MP3 = 1;                                         // reset to first 
      }
    } // end Play_Evil
    
};   // END class NeoPatterns


// Main Declarations of Objects

void Ring1Complete();

// Declare a NeoPatterns object and pass number of rings and pin#
NeoPatterns Ring1(12, 10, NEO_GRB + NEO_KHZ800, &Ring1Complete);

// SETUP
// Initialize everything and prepare to start
void setup()
{
  // Setup the Audio with MusicMaker
  //Serial.begin(115200);
  if (! musicPlayer.begin()) { // initialise the music player
     //Serial.println(F("Couldn't find VS1053, do you have the right pins defined?"));
     while (1);
  }
   //Serial.println(F("VS1053 found"));                                 // testing

  // Make a tone to indicate VS1053 is working
  musicPlayer.sineTest(0x44, 500);    
  
  if (!SD.begin(CARDCS_Pin)) {
    //Serial.println(F("SD failed, or not present"));                   // testing
    while (1);  // don't do anything more
  }
  //Serial.println("SD OK!");                                           // testing
  
  // list files - npo need to list the files
  //printDirectory(SD.open("/"), 0);                                    // testing
  
  // Set volume for left, right channels. 
  // Lower numbers = louder volume!
  musicPlayer.setVolume(30,30);                                         // Set Volume
  musicPlayer.useInterrupt(VS1053_FILEPLAYER_PIN_INT);                  // DREQ int
  musicPlayer.startPlayingFile("/track001.mp3");                        // test track
  musicPlayer.setVolume(Volume, Volume);                                // Set Volume
  
  pinMode (Flashlight_Pin, OUTPUT);                                     // Flashlight
  pinMode (Bluetooth_Pin, OUTPUT);                                      // Pin for powering the BlueToothe Module

  // Initialize all the pixelStrips
  Ring1.begin();

  // Activate patterns
  //Ring1.TheaterChase(Ring1.Color(255, 0, 125), Ring1.Color(0, 0, 50), 1000);

  Ring1.setBrightness(Brightness);                                      // Set the overall brightness.  - AYH

  CurrentPattern = 1;                                             //
  Ring1.setBrightness(50);                                              // 
  Ring1.WhiteLight();                                                   // 
  //Ring1.RainbowCycle(30);                                             // set speed the colours swirls/rotate

  // This section attaches callbacks to each declared Switch.
  // These include the click event, doubl-click and a long-press events.
  // There is also click events the start and stop on the long_press event.

  // Function Switch
  btn_Function.attachClick(btn_Function_click);                         // calls btn_Funtion_Click()
  btn_Function.attachDoubleClick(btn_Function_doubleclick);             // 
  btn_Function.attachLongPressStart(btn_Function_longPressStart);       //
  btn_Function.attachLongPressStop(btn_Function_longPressStop);         // 
  btn_Function.attachDuringLongPress(btn_Function_longPress);           // 

  // The Toggle Switch
  btn_Toggle.attachClick(btn_Toggle_click);
  btn_Toggle.attachDoubleClick(btn_Toggle_doubleclick);
  btn_Toggle.attachLongPressStart(btn_Toggle_longPressStart);
  btn_Toggle.attachLongPressStop(btn_Toggle_longPressStop);
  btn_Toggle.attachDuringLongPress(btn_Toggle_longPress);

  // The Increment Switch
  btn_Increment.attachClick(btn_Increment_click);
  btn_Increment.attachDoubleClick(btn_Increment_doubleclick);
  btn_Increment.attachLongPressStart(btn_Increment_longPressStart);
  btn_Increment.attachLongPressStop(btn_Increment_longPressStop);
  btn_Increment.attachDuringLongPress(btn_Increment_longPress);

  // The Decrement Switch
  btn_Decrement.attachClick(btn_Decrement_click);
  btn_Decrement.attachDoubleClick(btn_Decrement_doubleclick);
  btn_Decrement.attachLongPressStart(btn_Decrement_longPressStart);
  btn_Decrement.attachLongPressStop(btn_Decrement_longPressStop);
  btn_Decrement.attachDuringLongPress(btn_Decrement_longPress);

} // end setup

// Main loop
void loop()
{
  //CheckPlayer();                                                // 

  Ring1.Update();                                                 // Call the update method.
  //
  // Check for a Switch Press.
  // Note:  Although using interrupts would be a faster and more efficient way to handle the Switch press
  //        I Suspect there may be issues with interrupts and the neo_Pixel library, especially with
  //        the Trinket mini which has few ports. The OneButton library also allows other click events
  btn_Function.tick();                                            // calls tick() method of OneButton
  btn_Toggle.tick();                                              // When a click is detected the
  btn_Increment.tick();                                           // appropriate event is called
  btn_Decrement.tick();                                           // Click events are near end of code.
  
} // end Loop


/// File listing helper
void printDirectory(File dir, int numTabs) {
   while(true) {
     
     File entry =  dir.openNextFile();
     if (! entry) {
       // no more files
       //Serial.println("**nomorefiles**");
       break;
     }
     for (uint8_t i=0; i<numTabs; i++) {
       Serial.print('\t');
     }
     Serial.print(entry.name());
     if (entry.isDirectory()) {
       Serial.println("/");
       printDirectory(entry, numTabs+1);
     } else {
       // files have sizes, directories do not
       Serial.print("\t\t");
       Serial.println(entry.size(), DEC);
     }
     entry.close();
   }
  }
  
//------------------------------------------------------------
//Completion Routines - get called on completion of a pattern
//------------------------------------------------------------

// Ring1 Completion Callback
void Ring1Complete()
{

}


// OneButtom library Events
//  - Function Switch
//  - Toggle Switch
//  - Increment Switch
//  - Decrement Switch
//
//  Each Switch has the following events attached to it.
//    - Click, Double-Click, LongPressStart, LongPressStop, LongPress
//
// **************************************************************************
// Function Switch Events
// All the function button events do same thing no matter which pattern
void btn_Function_click()
{
  Ring1.NextPattern();                                          // Increment the Pattern
} 
void btn_Function_doubleclick()
{
  musicPlayer.startPlayingFile("/track002.mp3"); //tequila
  //musicPlayer.startPlayingFile("/rocklob.mp3"); // rock lobster 
} //
void btn_Function_longPressStart()    //once
{
  //
} 
void btn_Function_longPress()
{
// musicPlayer.startPlayingFile("/track001.mp3");      // cant do this, keeps trying to restart every cpu tick

} //
void btn_Function_longPressStop()   // once
{
  //
  // TEst Routine to scare crap out people, aka SNP
  //
   //digitalWrite(Flashlight_Pin, HIGH); 
   musicPlayer.stopPlaying();
   musicPlayer.setVolume(10,10);                          // crank it up. 
   //musicPlayer.startPlayingFile("/Track101.mp3");        // thunder 1 needs FULL battery or reboot
   musicPlayer.startPlayingFile("/Wind_1a.mp3");          // try this.
   
   digitalWrite(Flashlight_Pin, LOW);                     // ensure center light is off.

   
   
} //


// **************************************************************************
// Toggle Switch Events

void btn_Toggle_click()
{
  // determine the current pattern and act accordingly.
  switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            Ring1.Next_Colour();                                      // next colour
            break;
          case 2:       // Rainbow
            musicPlayer.setVolume(Volume,Volume);                          // crank it up. 
            musicPlayer.startPlayingFile("/track004.mp3");          // test scarbough
            break;
          case 3:       // Fade
            //
            musicPlayer.stopPlaying();
            musicPlayer.setVolume(Volume,Volume);                          // crank it up. 
            musicPlayer.startPlayingFile("/Feed_Me.mp3");          // 
            break;
          case 4:       // Sparks            
            break;
          case 5:       // Detect Evil
            break;
          case 6:
            break;
          case 7:
          Ring1.Next_Colour();          // note: currentl only 5 active patterns so this will never fire.   // next colour
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
  
}
void btn_Toggle_doubleclick()
{
  // switch to Detct Evil Pattern
  CurrentPattern = 5; 
  Ring1.DetectEvil(5);                              // Initialize 
  Ring1.DetectEvilUpdate();                         //                                         // // 
} //
void btn_Toggle_longPressStart()    //once
{
 
} //
void btn_Toggle_longPressStop()   // once
{
   // fire thunder
    musicPlayer.stopPlaying();
    musicPlayer.setVolume(1,1);                           // crank it up. 
    musicPlayer.startPlayingFile("/tundrBst.mp3"); 
   
} //
void btn_Toggle_longPress()
{
  // switch to Detect Evil/Good PAttern
  // test
  
} //
// **************************************************************************
// Increment Switch Events

void btn_Increment_click()
{
  switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            Ring1.More_Bright();          // increase brightness
            break;
          case 2:       // Rainbow
            Ring1.More_Bright();
            break;
          case 3:       // Fade
            Ring1.More_Bright();
            break;
          case 4:       // Sparks            
            Ring1.More_Bright();
            break;
          case 5:       // Detect Evil
            Ring1.More_Evil();              // run More_Evil function to fade to the evil colour (Red)
            // Ring1.More_Bright();
            break;
          case 6:       // 
            Ring1.More_Bright();
            break;
          case 7:       // 
            // adjust Volume
            Ring1.More_Volume();
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
  
} //
void btn_Increment_doubleclick()
{
// test for and act on CurrentPattern
   switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            break;
          case 2:       // Rainbow
                         Ring1.SpeedUp(); // change speed
                         //Ring1.Interval = Ring1.Interval - Interval_Increment;   // change pixel refresh rate by variable Interval_Increment
            break;
          case 3:       // Fade
            // increse speed,( shorten interval)
            Ring1.SpeedUp();
            break;
          case 4:       // Sparks            
            break;
          case 5:       // Detect Evil
            // go full bright
            Brightness = maxBrightness;
            Ring1.setBrightness(Brightness);
            break;
          case 6:       // 
            break;
          case 7:
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 

} //
void btn_Increment_longPressStart()    //once
{
   switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            digitalWrite(Flashlight_Pin, HIGH);     // activate the Center LED Array.
            break;
          case 2:       // Rainbow
            digitalWrite(Flashlight_Pin, HIGH);     // activate the Center LED Array.
            break;
          case 3:       // Fade
            digitalWrite(Flashlight_Pin, HIGH);     // activate the Center LED Array.
            break;
          case 4:       // Sparks - Setup pattern           
            digitalWrite(Bluetooth_Pin, HIGH);      // Activate the BlueTooth Module While Pressed
            break;
          case 5:       // Detect Evil
            //
            //Ring1.More_Evil();
            break;
          case 6:       //                         // Increase the Evil Colour
            break;
          case 7:
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
  
} //
void btn_Increment_longPressStop()   // once
{
   switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            digitalWrite(Flashlight_Pin, LOW);     // activate the Center LED Array.
            break;
          case 2:       // Rainbow
                        // change speed
            digitalWrite(Flashlight_Pin, LOW);     // activate the Center LED Array.
            // Ring1.Interval = Ring1.Interval - Interval_Increment;
            break;
          case 3:       // Fade
            digitalWrite(Flashlight_Pin, LOW);     // activate the Center LED Array.
            break;
          case 4:       // Sparks            
            digitalWrite(Bluetooth_Pin, LOW);      // Activate the BlueTooth Module While Pressed
            break;
          case 5:       // Detect Evil
            Ring1.Play_Evil();
            // digitalWrite(Flashlight_Pin, LOW);     // activate the Center LED Array.
            break;
          case 6:       // 
            break;
          case 7:
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
     
} //
void btn_Increment_longPress()
{
  // When is this fired? Once, while pressed?
} 

// **************************************************************************
// Decrement Switch Events

void btn_Decrement_click()
{
  // determine the current pattern and act accordingly.
  switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            Ring1.Less_Bright();
            break;
          case 2:       // Rainbow
            Ring1.Less_Bright();
            break;
          case 3:       // Fade
            Ring1.Less_Bright();
            break;
          case 4:       // Sparks            
            Ring1.Less_Bright();
            break;
          case 5:       // Detect Evil
            Ring1.Less_Evil();
            // Ring1.Less_Bright();
            break;
          case 6:       // 
            Ring1.Less_Bright();
            break;
          case 7:       // 
            // adjust Volume
            Ring1.Less_Volume();
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
} //
void btn_Decrement_doubleclick()
{
switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            
            break;
          case 2:       // Rainbow
            Ring1.SlowDown();            // change speed
            // Ring1.Interval = Ring1.Interval + Interval_Increment;  // better to call the function above
            break;
          case 3:       // Fade
            // Slow down the pattern
            Ring1.SlowDown();
            break;
          case 4:       // Sparks            
            break;
          case 5:       // Detect Evil
            break;
          case 6:       //            
            break;
          case 7:
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
  
} //
void btn_Decrement_longPressStart()    //once
{
  switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            break;
          case 2:       // Rainbow
            break;
          case 3:       // Fade
            //             
            break;
          case 4:       // Sparks            
            break;
          case 5:       // Detect Evil
            //Ring1.Less_Evil();
            break;
          case 6:       //            
            break;
          case 7:
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
  
} //
void btn_Decrement_longPressStop()   // once
{
  //
    switch (CurrentPattern)               //
        {
          case 1:       // Basic Light
            musicPlayer.stopPlaying();
            musicPlayer.setVolume(Volume,Volume);                          // crank it up. 
            musicPlayer.startPlayingFile("/Texan_3.mp3");          // 
            break;
          case 2:       // Rainbow
            musicPlayer.stopPlaying();
            musicPlayer.setVolume(Volume,Volume);                          // crank it up. 
            musicPlayer.startPlayingFile("/Country1.mp3");          // 
            
            break;
          case 3:       // Fade
            //             
            musicPlayer.stopPlaying();
            musicPlayer.setVolume(Volume,Volume);                          // crank it up. 
            musicPlayer.startPlayingFile("/NeilMix1.mp3");          // 
            
            break;
          case 4:       // Sparks            
            break;
          case 5:       // Detect Evil
            musicPlayer.stopPlaying();
            Ring1.Play_Good();          // try to play a good sound
            //Ring1.Less_Evil();
            break;
          case 6:       //            
            break;
          case 7:
            break;
          case 8:                               // Couple of extras for new patterns
            break;
          case 9:
            break;
          default:
            break;
        }   // end case 
}
void btn_Decrement_longPress()
{
  //
   
}



// **************************************************************************
// Testing 

  void CheckPlayer()
  {
    // test to see if player is ok and staus of current song
    // maybe check for random playing, sequential, playlist...

    //Serial.println("CheckPlayer");
    // for now go to next song
    if (musicPlayer.stopped()) 
    {
      // get a new song/
      GetNextSong(SD.open("/"));
    }
  }
  void GetNextSong(File dir)
  {

    //
    //musicPlayer.stopPlaying();
    
    // test for and stats playing the next song
     File entry =  dir.openNextFile();
     if (! entry) {
       // no more files

     }
     //Serial.print(entry.name());
     if (entry.isDirectory()) {
       // skip it
     } else 
     {      
       musicPlayer.startPlayingFile(entry.name());
     }
     // Serial.println("");
     // Serial.print("GetingNextSong:");
     // Serial.print(entry.name());

     entry.close();
  }

   void PlayError(int ErrorCode)
   {
    // test 
    if (ErrorCode = 1)
    {
      // 
     Serial.println("");
     Serial.println("Error");
      musicPlayer.sineTest(0x44, 550);
      delay(250);
      musicPlayer.sineTest(0x44, 250);
    }
   }

  // test 
   void FlashCurrentPattern()
   {
    for (int x=1;x=FlashCurrentPattern;x++)
    {
      digitalWrite(Flashlight_Pin, HIGH);     // activate the Center LED Array.
      delay (50);
      digitalWrite(Flashlight_Pin, LOW);     // activate the Center LED Array.
      delay (50);
    }
   }
